def compare_dicts(obj1, obj2):
    """Compares two objects recursively, handling dicts and lists.
    returns True if the objects are equal, False otherwise.
    """
    if obj1 == obj2:
        return True
    if isinstance(obj1, dict) and isinstance(obj2, dict):
        if obj1.keys() != obj2.keys():
            return False
        for key in obj1:
            if not compare_dicts(obj1[key], obj2[key]):
                return False
        return True
    if isinstance(obj1, list) and isinstance(obj2, list):
        if len(obj1) != len(obj2):
            return False
        for item1, item2 in zip(obj1, obj2):
            if not compare_dicts(item1, item2):
                return False
        return True
    return False



from __future__ import annotations
import logging
import traceback

from abc import ABC, abstractmethod
from typing import Any, Type

import dash_mantine_components as dmc
from dash import dcc, html
from dash_iconify import DashIconify


class Card(ABC):
    """Class to represent a card on the dashboard. This is an abstract class.

    This is the parent class for all cards on the dashboard. The following methods
    must be implemented in the child classes:
    - render: Render the card.
    - render_settings: Render the settings for the card.

    The following attributes should be set in the child classes. These are
    used to display the card on in the card gallery:
    - title: The title of the card.
    - description: The description of the card.
    - icon: The icon for the card.
    """

    title: str = "Test Card"
    description: str = "This is a sample card."
    icon = "mdi:file-document-edit"
    color: str = "blue"
    interval: int | None = None
    grid_settings: dict[str, int] | None = None
    debug = False  # Set this to True to display full error traceback on card

    def __init__(
        self,
        card_id,
        global_settings: dict[str, str] | None = None,
        card_settings: dict[str, str] | None = None,
    ):
        """Initialize the card.

        Args:
            card_id: The id of the card.
            global_settings: The global settings for the dashboard.
            card_settings: The settings for the card.
                These settings will be saved in a global store and will be passed
                into the CardManager which inturn passes it onto the card before
                it gets rendered.
        """
        self.id = card_id
        self.global_settings = global_settings or {}
        self.settings = card_settings or {}

    @abstractmethod
    def render(self):
        """Render the card.

        This method should return a Dash component that represents the card in
        the dashboard.
        """
        pass

    def render_container(self):
        """Renders a card with a menu on the top right corner.

        Returns:
            dash.html.Div: The card with a menu at the top.
        """
        buttons = html.Div(
            dmc.Menu(
                [
                    dmc.MenuTarget(
                        dmc.ActionIcon(
                            DashIconify(icon="material-symbols:more-horiz"),
                            size="xs",
                            radius="xl",
                            variant="light",
                            color="grey",
                        )
                    ),
                    dmc.MenuDropdown(
                        [
                            dmc.MenuItem(
                                "Settings",
                                id={"type": "card-settings", "index": self.id},
                                className="no-drag",
                                leftSection=DashIconify(icon="mdi:cog-outline"),
                            ),
                            dmc.MenuItem(
                                "Duplicate",
                                id={"type": "card-duplicate", "index": self.id},
                                className="no-drag",
                                leftSection=DashIconify(icon="mdi:content-copy"),
                            ),
                            dmc.MenuItem(
                                "Delete",
                                id={"type": "card-delete", "index": self.id},
                                className="no-drag",
                                leftSection=DashIconify(icon="mdi:trash-can-outline"),
                                c="red",
                            ),
                        ]
                    ),
                ],
            ),
            id={"type": "card-menu", "index": self.id},
            className="no-drag card-menu",
        )
        try:
            card_content = self.render()
        except Exception as e:
            logging.error(f"Error rendering card {self.id}: {str(e)}")
            logging.error(traceback.format_exc())
            if self.debug:
                card_content = html.Div(
                    html.Pre(
                        f"Error rendering card: {str(e)}\n{traceback.format_exc()}",
                    ),
                    style={
                        "color": "red",
                        "width": "100%",
                        "height": "100%",
                        "overflow": "auto",
                    },
                )
            else:
                card_content = dmc.Alert(
                    dmc.Text(f"Error rendering card: {str(e)}", ff="Consolas"),
                    color="red",
                    title="Error",
                    h="100%",
                )
        children: list[Any] = [
            dcc.Loading(
                html.Div(
                    id={"type": "card-content", "index": self.id},
                    style={"height": "100%"},
                    children=card_content,
                ),
                parent_style={"height": "100%"},
            ),
            buttons,
        ]
        if self.interval:
            children.append(
                dcc.Interval(
                    id={"type": "card-interval", "index": self.id},
                    interval=self.interval,
                )
            )

        return html.Div(
            children,
            style={"position": "relative", "height": "100%"},
            id=self.id,
        )

    def render_settings(self):
        """Render the settings for the card.

        This method should return a Dash component that represents the settings
        for the card in the settings modal. The settings drawer is displayed when
        the user clicks on the settings icon for the card.

        Each control in the settings modal should be a dash input field which has
        a value prop. The control should have an id that matches with
        the following template.
        `id={"type": "card-settings", "id": self.id, "setting": "control-id"}`
        Here `control-id` is the id of the control in the settings drawer.
        This control-id will be the key in the settings dictionary that is passed
        to the card's render method.

        Note: The control's value property is used to update the settings dictionary.
        Right now, no other property name is supported.
        """
        return dmc.Text("Settings not implemented yet.")

class GlobalSettings(ABC):
    """Class to represent the global settings for the dashboard. This is an abstract class.

    This is the parent class for all global settings on the dashboard. The following methods
    must be implemented in the child classes:
    - render: Render the global settings.

    The following attributes should be set in the child classes. These are
    used to display the global settings on in the card gallery:
    - title: The title of the global settings.
    - description: The description of the global settings.
    - icon: The icon for the global settings.
    """

    title: str = "Global Settings"
    description: str = "These settings apply to all cards on the dashboard."
    icon = "mdi:cog"

    def __init__(self, settings: dict[str, str] | None = None) -> None:
        """Initialize the global settings.

        Args:
            global_settings: The global settings for the dashboard.
        """
        self.settings: dict[str, str] = settings or {}

    @abstractmethod
    def render_settings(self):
        """Render the global settings.

        This method should return a Dash component that represents the global settings in
        the dashboard.
        """
        pass

class CardManager:
    """Class to manage the cards on the dashboard."""

    def __init__(self) -> None:
        self.card_classes: dict[str, Type[Card]] = {}
        self.global_settings_class: Type[GlobalSettings] | None = None

    def card_objects(
        self,
        card_config: dict[str, dict[str, Any]],
        global_settings: dict[str, str] | None = None,
    ) -> dict[str, Card]:
        card_config = card_config or {}
        global_settings = global_settings or {}
        cards: dict[str, Card] = {}
        for card_id, card_settings in card_config.items():
            card_class = card_settings.get("card_class")
            if not card_class or card_class not in self.card_classes:
                continue
            card = self.card_classes[card_class](
                card_id, global_settings, card_settings.get("settings", {})
            )
            cards[card_id] = card
        return cards

    def render(
        self,
        card_config: dict[str, dict[str, Any]],
        global_settings: dict[str, str] | None = None,
        debug=False,
    ) -> list[html.Div]:
        cards = self.card_objects(card_config, global_settings)
        for card in cards.values():
            card.debug = debug
        return [card.render_container() for card in cards.values()]

    def register_card_class(self, card_class: Type[Card]) -> None:
        """Register a card class with the card manager.

        Args:
            card_class: The class of the card to be registered.
        """
        self.card_classes[card_class.__name__] = card_class

    def register_global_settings_class(
        self, global_settings_class: Type[GlobalSettings]
    ) -> None:
        """Register a global settings class with the card manager.

        Args:
            global_settings_class: The class of the global settings to be registered.
        """
        self.global_settings_class = global_settings_class



from __future__ import annotations

import dash_mantine_components as dmc
from dash import dcc, html
from dash_iconify import DashIconify
from dash_snap_grid import DraggableDiv


def button_with_tooltip(id, icon, title, tooltip, **button_settings):
    return dmc.Tooltip(
        dmc.Button(
            id=id,
            leftSection=DashIconify(icon=icon),
            children=title,
            **button_settings,
        ),
        label=tooltip,
    )


def icon_with_tooltip(id, icon, title, tooltip, **button_settings):
    return dmc.ActionIcon(
        children=dmc.Tooltip(
            children=DashIconify(icon=icon, width=20, style={"margin": "5px"}),
            label=tooltip,
        ),
        id=id,
        **button_settings,
    )


def main_buttons(global_settings: bool = False):
    button_settings = {
        "size": "compact-s",
        "p": 4,
        "variant": "light",
    }
    children = []
    if global_settings:
        children.append(
            icon_with_tooltip(
                id="open-global-settings",
                icon="mdi:cog-outline",
                title="Settings",
                tooltip="Open the global settings dialog",
                **button_settings,
            )
        )
    children.extend(
        [
            icon_with_tooltip(
                id="add-cards",
                icon="mdi:plus",
                title="Add Cards",
                tooltip="Add new cards to the layout",
                **button_settings,
            ),
            dmc.ActionIconGroup(
                children=[
                    dcc.Upload(
                        id="upload-layout",
                        children=icon_with_tooltip(
                            id="upload-layout-button",
                            icon="mdi:upload",
                            title="Upload Layout",
                            tooltip="Upload a JSON file to load a layout.",
                            **button_settings,
                        ),
                    ),
                    icon_with_tooltip(
                        id="download-layout",
                        icon="mdi:download",
                        title="Download Layout",
                        tooltip="Download the current layout as a JSON file.",
                        **button_settings,
                    ),
                    icon_with_tooltip(
                        id="save-layout",
                        icon="mdi:content-save",
                        title="Save Layout",
                        tooltip="Save the current layout to the browser's local storage.",
                        **button_settings,
                    ),
                    icon_with_tooltip(
                        id="restore-layout",
                        icon="mdi:restore-clock",
                        title="Reset Layout",
                        tooltip="Restore the layout to last saved layout.",
                        **button_settings,
                    ),
                    icon_with_tooltip(
                        id="reset-layout",
                        icon="mdi:restore",
                        title="Reset Layout",
                        tooltip="Reset the layout to the default layout.",
                        color="yellow",
                        **button_settings,
                    ),
                    icon_with_tooltip(
                        id="clear-layout",
                        icon="mdi:delete",
                        title="Clear Layout",
                        tooltip="Clear all cards from the layout.",
                        color="red",
                        **button_settings,
                    ),
                ]
            ),
            dmc.Tooltip(
                dmc.Switch(
                    id="edit-layout",
                    offLabel=DashIconify(icon="material-symbols:edit-off", width=20),
                    onLabel=DashIconify(icon="material-symbols:edit", width=20),
                    size="md",
                    checked=False,
                    persistence=True,
                ),
                label="Toggle edit mode to modify, remove or move cards",
            ),
            dmc.Switch(
                id="color-scheme-toggle",
                offLabel=DashIconify(icon="radix-icons:moon", width=20),
                onLabel=DashIconify(icon="radix-icons:sun", width=20),
                size="md",
                persistence=True,
                ms="auto",
            ),
        ]
    )

    return dmc.Group(
        children,
        id="toolbar",
        p="xs",
    )


def get_title_layout(title: str, subtitle: str | None = None, logo: str | None = None):
    """Returns a layout for the title of the app.

    Args:
        title (str): The title of the app.
        subtitle (str): The subtitle of the app.
        logo (str): URL of the logo.

    Returns:
        dmc.Group: The title layout
    """
    items = []
    title_subtitle = []
    if logo:
        items.append(
            html.Img(src=logo, height=70),
        )
    title_subtitle.append(dmc.Title(title, order=2, c="blue"))
    if subtitle:
        title_subtitle.append(
            dmc.Text(
                subtitle,
                fw=300,
                fz="s",
                c="grey",
            )
        )
    items.append(
        dmc.Stack(
            title_subtitle,
            gap=0,
        )
    )
    return dmc.Group(items)


def render_card_preview(card_class) -> DraggableDiv:
    """Renders a card preview in the card gallery

    Args:
        card_class (Card): The card class to render.

    Returns:
        DraggableDiv: A draggable div with the card preview
    """
    return DraggableDiv(
        [
            dmc.Card(
                dmc.Group(
                    [
                        dmc.Paper(
                            dmc.ThemeIcon(
                                size="xl",
                                color=card_class.color,
                                variant="filled",
                                children=DashIconify(
                                    icon=card_class.icon,
                                    width=25,
                                ),
                            )
                        ),
                        dmc.Stack(
                            [
                                dmc.Text(card_class.title, fw=500, fz=20, c="gray"),
                                dmc.Text(card_class.description, fz=14, c="gray"),
                            ],
                            gap=0,
                        ),
                    ],
                    wrap="nowrap",
                ),
                style={"cursor": "grab"},
                p="sm",
                shadow="xs",
                withBorder=True,
            )
        ],
        id=card_class.__name__,
    )



from cardcanvas import CardCanvas, Card
import dash_mantine_components as dmc

settings = {
    "title": "CardCanvas Demo",
    "subtitle": "A Demo application showing the capabilities of CardCanvas",
    "start_config": {},
    "logo": "https://img.icons8.com/?size=80&id=cjlQopC5NR3D&format=png",
    "grid_compact_type": "vertical",
    "grid_row_height": 100,
}


class TextCard(Card):
    title = "White text with a background color"
    description = "Testing out CardCanvas"
    icon = "mdi:file-document-edit"

    def render(self):
        return dmc.Card(
            dmc.Title(
                self.settings.get("text", "Hello CardCanvas"),
                c="white",
            ),
            bg=self.settings.get("color", "blue"),
            style={"height": "100%", "width": "100%"},
        )

    def render_settings(self):
        return dmc.Stack(
            [
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "text"},
                    value=self.settings.get("text", "Hello CardCanvas"),
                ),
                dmc.ColorPicker(
                    id={"type": "card-settings", "id": self.id, "setting": "color"},
                    value=self.settings.get("color", "grey"),
                ),
            ]
        )


canvas = CardCanvas(settings)
canvas.card_manager.register_card_class(TextCard)

canvas.app.run_server(debug=True)



from dash import Dash
from cardcanvas import CardCanvas, Card


class TestCard(Card):
    def render(self):
        return "Hello, World!"


def test_main():
    settings = {
        "title": "My Dash App",
        "start_config": {},
    }
    dashboard = CardCanvas(settings)
    dashboard.card_manager.register_card_class(TestCard)
    assert dashboard.card_manager.card_classes == {"TestCard": TestCard}
    assert isinstance(dashboard.app, Dash)



import base64
import copy
import json
import logging
from typing import Any
from uuid import uuid4

import dash_mantine_components as dmc
from dash import (
    ALL,
    MATCH,
    Dash,
    Input,
    Output,
    State,
    ctx,
    dcc,
    html,
    no_update,
    _dash_renderer,
)
from dash_snap_grid import ResponsiveGrid
from dash_iconify import DashIconify

from . import ui
from .card_manager import CardManager
from .settings import DEFAULT_THEME

_dash_renderer._set_react_version("18.2.0")
dmc.add_figure_templates()


class CardCanvas:
    def __init__(
        self, settings: dict[str, Any], dash_options: dict[str, Any] | None = None
    ):
        self.settings = settings
        self.card_manager = CardManager()
        self.dash_options = dash_options or {}

    def run(self):
        self.app.run_server(debug=True)

    @property
    def app(self):
        if not hasattr(self, "_app"):
            self._app = self._create_app(self.settings)
        return self._app

    def _create_app(self, settings: dict[str, Any]) -> Dash:
        title = settings.get("title", "Card Canvas")
        subtitle = settings.get("subtitle", None)
        title_component = settings.get("title_component", None)
        footer_component = settings.get("footer_component", None)
        start_config = settings.get("start_config", {})
        start_card_config = start_config.get("card_config", {})
        start_card_layout = start_config.get("card_layouts", {"lg": []})
        start_global_settings = start_config.get("global_settings", {})
        logo = settings.get("logo", None)
        theme = settings.get("theme", DEFAULT_THEME)

        show_global_settings = settings.get("show_global_settings", True)
        app = Dash(
            __name__,
            **self.dash_options,
            external_stylesheets=[dmc.styles.NOTIFICATIONS, dmc.styles.CHARTS],
            suppress_callback_exceptions=True,
        )
        app.title = f"{title}: {subtitle}" if subtitle else title

        title_layout = dmc.Group(
            [
                title_component
                if title_component
                else ui.get_title_layout(title, subtitle=subtitle, logo=logo),
                dmc.ActionIcon(
                    id="open-main-menu",
                    children=DashIconify(icon="mdi:menu"),
                    variant="outline",
                ),
            ],
            justify="space-between",
            p="xs",
        )

        background_color = settings.get("background_color")

        main_buttons = dmc.Collapse(
            id="main-menu-collapse",
            children=[ui.main_buttons(global_settings=show_global_settings)],
            opened=True,
            style={
                "position": "sticky",
                "top": 0,
                "zIndex": 10,
                "backgroundColor": background_color,
            },
        )

        stage_children = [
            title_layout,
            main_buttons,
            ResponsiveGrid(
                id="card-grid",
                children=[],
                cols=settings.get(
                    "grid_cols",
                    {"xl": 24, "lg": 18, "md": 12, "sm": 6, "xs": 4, "xxs": 2},
                ),
                breakpoints=settings.get(
                    "grid_breakpoints",
                    {
                        "xl": 1920,
                        "lg": 1200,
                        "md": 1080,
                        "sm": 768,
                        "xs": 576,
                        "xxs": 480,
                    },
                ),
                rowHeight=settings.get("grid_row_height", 50),
                compactType=settings.get("grid_compact_type", None),
                draggableCancel=".no-drag *",
                isDroppable=True,
                layouts={"lg": []},
                width=100,
            ),
        ]
        if footer_component:
            stage_children.append(footer_component)

        stage_layout = dmc.Container(
            fluid=True,
            children=stage_children,
            style={
                "backgroundColor": background_color,
                "minHeight": "100vh",
            },
        )

        invisible_controls = html.Div(
            children=[
                dcc.Store(id="cardcanvas-main-store", storage_type="local"),
                dcc.Store(
                    id="cardcanvas-config-store",
                    storage_type="memory",
                ),
                dcc.Store(
                    id="cardcanvas-layout-store",
                    storage_type="memory",
                ),
                dcc.Store(
                    id="cardcanvas-global-store",
                    storage_type="memory",
                ),
                dcc.Download(id="download-layout-data"),
                dmc.NotificationContainer(id="notification-container"),
            ],
        )

        settings_layout = dmc.Drawer(
            id="settings-layout",
            padding="md",
            closeOnClickOutside=False,
            withOverlay=False,
            position="right",
            lockScroll=False,
        )

        main_components = [stage_layout, settings_layout, invisible_controls]

        app.layout = dmc.MantineProvider(
            children=main_components,
            theme=theme,
            id="mantine-provider",
            forceColorScheme="light",
        )

        @app.callback(
            Output("cardcanvas-config-store", "data"),
            Output("cardcanvas-layout-store", "data"),
            Output("cardcanvas-global-store", "data"),
            Input(app.layout, "layout"),
            State("cardcanvas-main-store", "data"),
        )
        def load_layout(layout, main_store):
            if not main_store:
                main_store = {}

            card_config = main_store.get("card_config", start_card_config)
            card_layouts = main_store.get("card_layouts", start_card_layout)
            global_settings = main_store.get("global_settings", {})
            return card_config, card_layouts, global_settings

        @app.callback(
            Output("card-grid", "children"),
            Output("card-grid", "layouts"),
            Input("cardcanvas-config-store", "data"),
            Input("cardcanvas-layout-store", "data"),
            Input("cardcanvas-global-store", "data"),
            prevent_initial_call=True,
        )
        def load_cards(
            card_config_store,
            card_layout_store,
            global_settings,
        ):
            new_children = self.card_manager.render(
                card_config_store,
                global_settings=global_settings,
                debug=self.app.server.debug,
            )
            new_layout = card_layout_store
            return (
                new_children,
                new_layout,
            )

        @app.callback(
            Output("cardcanvas-main-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Output("notification-container", "sendNotifications"),
            Input("save-layout", "n_clicks"),
            State("card-grid", "layouts"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-global-store", "data"),
            prevent_initial_call=True,
        )
        def save_reset_cards(nclicks, card_layouts, card_config, global_settings):
            if not nclicks:
                return no_update, no_update, no_update
            return (
                {
                    "card_layouts": card_layouts,
                    "card_config": card_config,
                    "global_settings": global_settings,
                },
                # This is required since there may be changes in the layout which
                # are not reflected in the cardcanvas_layout_store
                card_layouts,
                [
                    dict(
                        title="Layout Saved",
                        message="The layout has been saved",
                        color="teal",
                        action="show",
                    )
                ],
            )

        @app.callback(
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-global-store", "data", allow_duplicate=True),
            Output("notification-container", "sendNotifications", allow_duplicate=True),
            Input("restore-layout", "n_clicks"),
            State("cardcanvas-main-store", "data"),
            prevent_initial_call=True,
        )
        def reset_layouts(nclicks, main_store):
            if not nclicks or not main_store or not isinstance(main_store, dict):
                return no_update, no_update, no_update
            return (
                main_store.get("card_layouts", start_card_layout),
                main_store.get("card_config", start_card_config),
                main_store.get("global_settings", start_global_settings),
                [
                    dict(
                        title="Layout Reset",
                        message="The layout has been reset to the last saved state",
                        color="orange",
                        action="show",
                    )
                ],
            )

        @app.callback(
            Output("settings-layout", "opened"),
            Output("settings-layout", "children"),
            Input("open-global-settings", "n_clicks"),
            State("cardcanvas-global-store", "data"),
            prevent_initial_call=True,
        )
        def open_settings(nclicks, global_settings):
            children = [
                dmc.Title("Global Settings", order=2),
                dmc.Text(
                    "These are the global settings. These apply to all the cards",
                    variant="muted",
                ),
            ]
            if show_global_settings and self.card_manager.global_settings_class:
                global_settings = self.card_manager.global_settings_class(
                    global_settings
                )
                children.extend(
                    [
                        global_settings.render_settings(),
                        dmc.Button("OK", id="global-settings-ok"),
                    ]
                )
            children = [dmc.Stack(children)]
            return True, children

        @app.callback(
            Output("cardcanvas-global-store", "data", allow_duplicate=True),
            Output("settings-layout", "opened", allow_duplicate=True),
            Input("global-settings-ok", "n_clicks"),
            State({"type": "global-settings", "setting": ALL}, "id"),
            State({"type": "global-settings", "setting": ALL}, "value"),
            State({"type": "global-settings", "setting": ALL}, "checked"),
            prevent_initial_call=True,
        )
        def save_global_settings(nclicks, ids, values, checked_values):
            if not nclicks or not ctx.triggered:
                return no_update, no_update
            global_settings = {}
            for idx, value, checked in zip(ids, values, checked_values):
                setting = idx.get("setting")
                if value is None and (checked in [True, False]):
                    value = checked
                global_settings[setting] = value
            return global_settings, False

        @app.callback(
            Output("settings-layout", "opened", allow_duplicate=True),
            Output("settings-layout", "children", allow_duplicate=True),
            Input("add-cards", "n_clicks"),
            prevent_initial_call=True,
        )
        def add_cards(nclicks):
            children = [
                dmc.Stack(
                    [
                        dmc.Title("Add Cards", order=2),
                        dmc.Text(
                            "These are the cards you can add to the dashboard."
                            " Drag and drop them on the grid where you want them to be"
                            " displayed. Configure them by clicking on the settings icon.",
                            variant="muted",
                        ),
                        dmc.Stack(
                            [
                                ui.render_card_preview(card_class)
                                for card_class in self.card_manager.card_classes.values()
                            ]
                        ),
                    ]
                )
            ]
            return True, children

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Input("card-grid", "droppedItem"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-layout-store", "data"),
            prevent_initial_call=True,
        )
        def add_new_card(dropped_item, card_config, card_layouts):
            if not dropped_item:
                return no_update, no_update
            card_id = str(uuid4())
            new_layout_item = {
                "i": card_id,
                "x": dropped_item["x"],
                "y": dropped_item["y"],
                "w": dropped_item["w"],
                "h": dropped_item["h"],
            }
            # dropped_item["i"] returns the id of dropped object. In this case, it is the card class.
            card_class = dropped_item["i"]
            card_config[card_id] = {"card_class": card_class, "settings": {}}
            card_class_obj = self.card_manager.card_classes.get(card_class)
            if (
                card_class_obj
                and hasattr(card_class_obj, "grid_settings")
                and isinstance(card_class_obj.grid_settings, dict)
            ):
                new_layout_item.update(card_class_obj.grid_settings)
            if not card_layouts:
                card_layouts = {"lg": []}
            for key in card_layouts.keys():
                card_layouts[key].append(new_layout_item)
            return card_config, card_layouts

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Input({"type": "card-duplicate", "index": ALL}, "n_clicks"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-layout-store", "data"),
            State("card-grid", "layout"),
            prevent_initial_call=True,
        )
        def duplicate_card(nclicks, card_config, card_layouts, card_layout):
            if not card_config:
                return no_update, no_update
            if not any(nclicks) or not ctx.triggered:
                return no_update, no_update
            if not ctx.triggered_id or not isinstance(ctx.triggered_id, dict):
                return no_update, no_update
            card_id = ctx.triggered_id.get("index")
            new_card_id = str(uuid4())
            new_card_layout = copy.deepcopy(
                next((item for item in card_layout if item["i"] == card_id))
            )
            new_card = copy.deepcopy(card_config.get(card_id, None))
            new_card["id"] = new_card_id
            new_card_layout["i"] = new_card_id
            card_config[new_card_id] = new_card
            for key in card_layouts.keys():
                card_layouts[key].append(new_card_layout)
            return card_config, card_layouts

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Input({"type": "card-delete", "index": ALL}, "n_clicks"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-layout-store", "data"),
            prevent_initial_call=True,
        )
        def delete_card(nclicks, card_config, card_layouts):
            if not card_config:
                return no_update, no_update
            if not any(nclicks) or not ctx.triggered:
                return no_update, no_update
            if not ctx.triggered_id or not isinstance(ctx.triggered_id, dict):
                return no_update, no_update
            card_id = ctx.triggered_id.get("index")
            card_config.pop(card_id, None)
            for key in card_layouts.keys():
                card_layouts[key] = [
                    item for item in card_layouts[key] if item["i"] != card_id
                ]
            return card_config, card_layouts

        @app.callback(
            Output("settings-layout", "children", allow_duplicate=True),
            Output("settings-layout", "opened", allow_duplicate=True),
            Input({"type": "card-settings", "index": ALL}, "n_clicks"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-global-store", "data"),
            prevent_initial_call=True,
        )
        def open_card_settings(nclicks, card_config, global_settings):
            if not any(nclicks) or not ctx.triggered or not ctx.triggered_id:
                return no_update, no_update
            if not card_config:
                card_config = start_card_config
            card_id = ctx.triggered_id.get("index")
            card_objects = self.card_manager.card_objects(card_config, global_settings)
            if card_id not in card_objects:
                return dmc.Alert(
                    f"Card with id: {card_id} not found", color="red"
                ), True
            card = card_objects[card_id]
            return dmc.Stack(
                [
                    dmc.Stack(
                        [
                            dmc.Text(card.title, fw="bold", size="lg", c="blue", mb=0),
                            dmc.Text(
                                f"Card ID: {card.id}",
                                fw="bold",
                                size="sm",
                                c="gray",
                            ),
                        ],
                        gap=2,
                    ),
                    card.render_settings(),
                    dmc.Button(
                        "OK",
                        id="card-settings-ok",
                    ),
                ]
            ), True

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("settings-layout", "opened", allow_duplicate=True),
            Input("card-settings-ok", "n_clicks"),
            State({"type": "card-settings", "id": ALL, "setting": ALL}, "id"),
            State({"type": "card-settings", "id": ALL, "setting": ALL}, "value"),
            State({"type": "card-settings", "id": ALL, "setting": ALL}, "checked"),
            State("cardcanvas-config-store", "data"),
            prevent_initial_call=True,
        )
        def save_card_settings(nclicks, ids, values, checked_values, card_config):
            if not nclicks or not ctx.triggered:
                return no_update, no_update
            for idx, value, checked in zip(ids, values, checked_values):
                card_id = idx.get("id")
                setting = idx.get("setting")
                if card_id not in card_config:
                    continue
                if value is None and (checked in [True, False]):
                    value = checked
                card_config[card_id]["settings"][setting] = value
            return card_config, False

        @app.callback(
            Output("card-grid", "isDraggable"),
            Output("card-grid", "isResizable"),
            Output({"type": "card-menu", "index": ALL}, "style"),
            Input({"type": "card-menu", "index": ALL}, "id"),
            Input("edit-layout", "checked"),
            prevent_initial_call=True,
        )
        def toggle_edit_mode(ids, checked):
            if checked:
                return True, True, [{"display": "block"}] * len(ids)
            return False, False, [{"display": "none"}] * len(ids)

        @app.callback(
            Output({"type": "card-content", "index": MATCH}, "children"),
            Input({"type": "card-interval", "index": MATCH}, "n_intervals"),
            State("cardcanvas-config-store", "data"),
            State("cardcanvas-global-store", "data"),
        )
        def update_card(n_intervals, cards_config, global_settings):
            if not ctx.triggered_id or not cards_config:
                return no_update
            card_objects = self.card_manager.card_objects(cards_config, global_settings)
            card_id = ctx.triggered_id.get("index")
            card = card_objects[card_id]
            return card.render()

        @app.callback(
            Output("download-layout-data", "data"),
            Input("download-layout", "n_clicks"),
            State("cardcanvas-main-store", "data"),
            prevent_initial_call=True,
        )
        def download_layout(nclicks, main_store):
            if not nclicks or not main_store:
                return no_update
            return dict(
                content=json.dumps(main_store), filename="layout.json", type="json"
            )

        @app.callback(
            Output("cardcanvas-main-store", "data", allow_duplicate=True),
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Input("upload-layout", "contents"),
            prevent_initial_call=True,
        )
        def upload_layout(contents):
            if not contents:
                return no_update
            try:
                content_type, content_string = contents.split(",")
                decoded = base64.b64decode(content_string)
                content = decoded.decode("utf-8")
                data = json.loads(content)
                return (
                    data,
                    data.get("card_config", start_card_config),
                    data.get("card_layouts", start_card_layout),
                )
            except Exception as e:
                logging.error(e)
            return {}

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Output("cardcanvas-global-store", "data", allow_duplicate=True),
            Output("notification-container", "sendNotifications", allow_duplicate=True),
            Input("clear-layout", "n_clicks"),
            prevent_initial_call=True,
        )
        def clear_layout(nclicks):
            if not nclicks:
                return no_update, no_update, no_update, no_update
            return (
                {},
                {},
                {},
                [
                    dict(
                        title="Layout Cleared",
                        message=(
                            "The layout has been cleared."
                            " Click on save to save the changes."
                            " Click on restore to restore the layout.",
                        ),
                        color="red",
                        action="show",
                    )
                ],
            )

        @app.callback(
            Output("cardcanvas-config-store", "data", allow_duplicate=True),
            Output("cardcanvas-layout-store", "data", allow_duplicate=True),
            Output("notification-container", "sendNotifications", allow_duplicate=True),
            Input("reset-layout", "n_clicks"),
            prevent_initial_call=True,
        )
        def reset_layout(nclicks):
            if not nclicks:
                return no_update, no_update, no_update
            return (
                start_card_config,
                start_card_layout,
                [
                    dict(
                        title="Layout Reset",
                        message=(
                            "The layout has been reset to default layout."
                            " Click on save to save the changes."
                            " Click on restore to restore the layout.",
                        ),
                        color="red",
                        action="show",
                    )
                ],
            )

        @app.callback(
            Output("main-menu-collapse", "opened"),
            Input("open-main-menu", "n_clicks"),
            State("main-menu-collapse", "opened"),
            prevent_initial_call=True,
        )
        def open_main_menu(n_clicks, status):
            return not status

        @app.callback(
            Output("mantine-provider", "forceColorScheme"),
            Input("color-scheme-toggle", "checked"),
            State("mantine-provider", "forceColorScheme"),
        )
        def switch_theme(toggle, theme):
            return "light" if toggle else "dark"

        return app



from pathlib import Path
import json
from cardcanvas import CardCanvas, Card
from dash import (
    html,
    dcc,
    callback,
    Input,
    State,
    Output,
    MATCH,
    ALL,
    Patch,
    callback_context,
    no_update,
)
import dash_mantine_components as dmc
import plotly.express as px
import plotly.io as pio
import pandas as pd
from dash_iconify import DashIconify

from data import nea_data as data

settings = {
    "title": "The National Endowment for the Arts Creative Writing Fellowships",
    "subtitle": "Writers sponsored by NEA fellowships from the organization’s founding in 1965 to 2024",
    "logo": "https://upload.wikimedia.org/wikipedia/commons/3/3b/National_Endowment_for_the_Arts_%28NEA%29_Logo_2018_Square_on_Black.svg",
    "start_config": json.loads((Path(__file__).parent / "layout.json").read_text()),
    "grid_compact_type": "vertical",
    "grid_row_height": 120,
    "show_global_settings": False,
}


class HistogramCard(Card):
    title = "Histogram"
    description = "This card shows a histogram of a given dataset"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}
    debug = True

    def render(self):
        column = self.settings.get("column", None)
        color = self.settings.get("color", None)
        nbins = self.settings.get("bins", 20)
        title = self.settings.get("title", "Histogram")
        description = self.settings.get("description", f"Histogram of {column}")

        figure = px.histogram(
            data,
            x=column,
            color=color,
            nbins=nbins,
            template="mantine_light",
        )
        figure.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        return dmc.Card(
            [
                dmc.Text(
                    title, fz="30px", fw=600, c="blue"
                ),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=figure,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        column = self.settings.get("column", None)
        color = self.settings.get("color", None)
        nbins = self.settings.get("bins", 20)
        title = self.settings.get("title", "Histogram")
        description = self.settings.get("description", "Histogram description")
        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "column",
                    },
                    label="Column",
                    value=column,
                    searchable=True,
                    # numeric columns in data
                    data=[
                        {"label": column, "value": column}
                        for column in data.select_dtypes(include="number").columns
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "color",
                    },
                    label="Color",
                    value=color,
                    searchable=True,
                    data=[
                        {"label": i, "value": i}
                        for i in data.select_dtypes(exclude="number").columns
                    ],
                ),
                dmc.NumberInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "bins",
                    },
                    label="Bins",
                    value=nbins,
                    min=1,
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )


def generate_filter(column: pd.Series, input_id, default_value=None):
    """Creating a filter based on the column type and it's unique values
    Used in heatmap card to filter the data based on the column values
    """
    column = column.dropna()
    card_id = input_id["id"]
    filter_type = input_id["setting"]
    if column.unique().shape[0] == 0:
        return dmc.Text("No data to filter", fz="14px", fw=600, c="red")
    elif column.unique().shape[0] == 1:
        return dmc.Text("Only one value, no need to filter", fz="14px", fw=600)
    if column.dtype in ["object", "string", "bool", "category"] or len(column.unique()) < 5:
        sorted_unique = sorted(column.unique().tolist())
        if len(sorted_unique) > 100:
            return dmc.Text(
                "Too many unique values to show filter", fz="14px", fw=600, c="red"
            )
        return [
            dmc.Text("Filter", fz="14px", fw=600),
            dmc.ScrollArea(
                dmc.CheckboxGroup(
                    id={
                        "type": "card-settings",
                        "id": card_id,
                        "setting": f"{filter_type}-filter",
                    },
                    value=default_value or sorted_unique,
                    children=dmc.Stack(
                        [dmc.Checkbox(label=str(x), value=str(x)) for x in sorted_unique]
                    ),
                ),
                style={"maxHeight": "250px", "overflowY": "auto"},
            ),
        ]
    return [
        dmc.Text("Filter", fz="14px", fw=600),
        dmc.RangeSlider(
            id={
                "type": "card-settings",
                "id": card_id,
                "setting": f"{filter_type}-filter",
            },
            value=default_value or [column.min(), column.max()],
            min=column.min(),
            max=column.max(),
            minRange=(column.max() - column.min()) / 100,
        ),
    ]


class HeatMap(Card):
    title = "Heatmap"
    description = "This card shows a heatmap of a given dataset"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        x = self.settings.get("x", None)
        x_filter = self.settings.get("x-filter", None)
        y = self.settings.get("y", None)
        y_filter = self.settings.get("y-filter", None)
        nbinsx = self.settings.get("nbinsx", 20)
        nbinsy = self.settings.get("nbinsy", 20)
        title = self.settings.get("title", "Heatmap")
        description = self.settings.get("description", f"Heatmap of {x} vs {y}")

        filtered_data = data.loc[:, [x, y]]
        if x_filter is not None:
            if filtered_data[x].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[x].isin(x_filter)]
            else:
                filtered_data = filtered_data[
                    (filtered_data[x] >= x_filter[0])
                    & (filtered_data[x] <= x_filter[1])
                ]
        if y_filter is not None:
            if filtered_data[y].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[y].isin(y_filter)]
            else:
                filtered_data = filtered_data[
                    (filtered_data[y] >= y_filter[0])
                    & (filtered_data[y] <= y_filter[1])
                ]
        figure = px.density_heatmap(
            filtered_data,
            x=x,
            y=y,
            nbinsx=nbinsx,
            nbinsy=nbinsy,
            template="mantine_light",
        )
        figure.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        return dmc.Card(
            [
                dmc.Text(
                    title, fz="30px", fw=600, c="blue"
                ),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=figure,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        x = self.settings.get("x", None)
        y = self.settings.get("y", None)
        x_filter = self.settings.get("x-filter", None)
        y_filter = self.settings.get("y-filter", None)
        nbinsx = self.settings.get("nbinsx", 20)
        nbinsy = self.settings.get("nbinsy", 20)
        title = self.settings.get("title", "Heatmap")
        description = self.settings.get("description", "Heatmap description")

        x_filter_children = None
        if x is not None:
            x_filter_children = generate_filter(
                data[x],
                {"type": "card-settings", "id": self.id, "setting": "x"},
                default_value=x_filter,
            )
        y_filter_children = None
        if y is not None:
            y_filter_children = generate_filter(
                data[y],
                {"type": "card-settings", "id": self.id, "setting": "y"},
                default_value=y_filter,
            )

        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "x",
                    },
                    label="X",
                    value=x,
                    searchable=True,
                    # numeric columns in data
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "x-filter",
                    },
                    children=x_filter_children,
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "y",
                    },
                    label="Y",
                    value=y,
                    searchable=True,
                    data=[{"label": i, "value": i} for i in data.columns],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "y-filter",
                    },
                    children=y_filter_children,
                ),
                dmc.NumberInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "nbinsx",
                    },
                    label="Number of bins in x direction",
                    value=nbinsx,
                    min=5,
                ),
                dmc.NumberInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "nbinsy",
                    },
                    label="Number of bins in y direction",
                    value=nbinsy,
                    min=5,
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )

    @callback(
        Output(
            {"type": "card-settings", "id": MATCH, "container": "x-filter"}, "children"
        ),
        Input({"type": "card-settings", "id": MATCH, "setting": "x"}, "value"),
    )
    def update_filter_x(value):
        """If the column is categorical, show a dropdown to filter the data
        else if data is numeric, show a slider to filter the data"""
        if value is None:
            return no_update
        column = data[value]
        # get the input id
        ctx = callback_context
        if not ctx.triggered_id:
            return no_update
        input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
        return generate_filter(column, input_id)

    @callback(
        Output(
            {"type": "card-settings", "id": MATCH, "container": "y-filter"}, "children"
        ),
        Input({"type": "card-settings", "id": MATCH, "setting": "y"}, "value"),
    )
    def update_filter_y(value):
        if value is None:
            return no_update
        column = data[value]
        ctx = callback_context
        if not ctx.triggered_id:
            return no_update
        input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
        return generate_filter(column, input_id)


class ViolinCard(Card):
    title = "Violin"
    description = "This card shows a violin plot of a given dataset"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        x = self.settings.get("x", None)
        y = self.settings.get("y", None)
        title = self.settings.get("title", "Violin plot")
        description = self.settings.get("description", f"Violin plot of {y} by {x}")
        fig = px.violin(
            data,
            x=x,
            y=y,
            template="mantine_light",
        )
        fig.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        fig.update_xaxes(
            categoryorder="array",
            categoryarray=data[x].unique(),
        )
        return dmc.Card(
            [
                dmc.Text(
                    title,
                    fz="30px",
                    fw=600,
                    c="blue",
                ),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=fig,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        x = self.settings.get("x", None)
        y = self.settings.get("y", None)
        title = self.settings.get("title", "Violin plot")
        description = self.settings.get("description", "Violin plot description")
        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "x",
                    },
                    label="X",
                    value=x,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "y",
                    },
                    label="Y",
                    value=y,
                    searchable=True,
                    # numeric columns in data
                    data=[
                        {"label": column, "value": column}
                        for column in data.select_dtypes(include="number").columns
                    ],
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )


class BarChartCard(Card):
    title = "Bar Chart"
    description = "This card shows a bar chart of a given dataset"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        x = self.settings.get("x", None)
        x_filter = self.settings.get("x-filter", None)
        y = self.settings.get("y", None)
        y_filter = self.settings.get("y-filter", None)
        color = self.settings.get("color", None)
        barmode = self.settings.get("barmode", "group")
        aggregation = self.settings.get("aggregation", "count")
        title = self.settings.get("title", "Bar Chart")
        description = self.settings.get("description", f"Bar chart of {y} by {x}")

        # apply filters
        filtered_data = data
        if x_filter is not None:
            if filtered_data[x].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[x].isin(x_filter)]
            else:
                filtered_data = filtered_data[
                    (filtered_data[x] >= x_filter[0])
                    & (filtered_data[x] <= x_filter[1])
                ]
        if y_filter is not None:
            if filtered_data[y].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[y].isin(y_filter)]
            else:
                filtered_data = filtered_data[
                    (filtered_data[y] >= y_filter[0])
                    & (filtered_data[y] <= y_filter[1])
                ]
        if color is None:
            if x and y:
                grouped_data = (
                    filtered_data.groupby(x)[y].agg(aggregation).reset_index()
                )
            else:
                grouped_data = pd.DataFrame()
        else:
            if x and color and y:
                grouped_data = (
                    filtered_data.groupby([x, color])[y].agg(aggregation).reset_index()
                )
            else:
                grouped_data = pd.DataFrame()

        fig = px.bar(
            template="mantine_light",
            data_frame=grouped_data,
            x=x,
            y=y,
            color=color,
            barmode=barmode
        )

        fig.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        return dmc.Card(
            [
                dmc.Text(
                    title,
                    fz="30px",
                    fw=600,
                    c="blue",
                ),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=fig,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        x = self.settings.get("x", None)
        x_filter = self.settings.get("x-filter", None)
        y = self.settings.get("y", None)
        y_filter = self.settings.get("y-filter", None)
        color = self.settings.get("color", None)
        barmode = self.settings.get("barmode", "group")
        aggregation = self.settings.get("aggregation", "sum")
        title = self.settings.get("title", "Bar Chart")
        description = self.settings.get("description", "Bar chart description")

        x_filter_children = None
        if x is not None:
            x_filter_children = generate_filter(
                data[x],
                {"type": "card-settings", "id": self.id, "setting": "x"},
                default_value=x_filter,
            )

        y_filter_children = None
        if y is not None:
            y_filter_children = generate_filter(
                data[y],
                {"type": "card-settings", "id": self.id, "setting": "y"},
                default_value=y_filter,
            )
        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "x",
                    },
                    label="X",
                    value=x,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "x-filter",
                    },
                    children=x_filter_children,
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "y",
                    },
                    label="Y",
                    value=y,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "y-filter",
                    },
                    children=y_filter_children,
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "aggregation",
                    },
                    label="Aggregation",
                    value=aggregation,
                    data=[
                        {"label": "Sum", "value": "sum"},
                        {"label": "Mean", "value": "mean"},
                        {"label": "Count", "value": "count"},
                        {"label": "Min", "value": "min"},
                        {"label": "Max", "value": "max"},
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "color",
                    },
                    label="Color",
                    value=color,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "barmode",
                    },
                    label="Bar Mode",
                    value=barmode,
                    data=[
                        {"label": "Grouped", "value": "group"},
                        {"label": "Stacked", "value": "stack"},
                    ],
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )


class TopNBarChartCard(Card):
    title = "Top Bar Chart"
    description = "Show the top n entries for a given column"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        column = self.settings.get("column", None)
        column_filter = self.settings.get("column-filter", None)
        n = self.settings.get("n", 10)
        title = self.settings.get("title", "Top N Bar Chart")
        description = self.settings.get("description", f"Top {n} entries of {column}")

        filtered_data = data
        if column_filter is not None:
            if filtered_data[column].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[column].isin(column_filter)]
            else:
                column_filter = [float(x) for x in column_filter]
                filtered_data = filtered_data[
                    (filtered_data[column] >= column_filter[0])
                    & (filtered_data[column] <= column_filter[1])
                ]

        top_n = filtered_data[column].value_counts().head(n).reset_index()
        top_n.columns = [column, "count"]
        fig = px.bar(
            top_n,
            x=column,
            y="count",
            template="mantine_light",
        )
        fig.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        return dmc.Card(
            [
                dmc.Text(
                    title,
                    fz="30px",
                    fw=600,
                    c="blue",
                ),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=fig,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        column = self.settings.get("column", None)
        column_filter = self.settings.get("column-filter", None)
        n = self.settings.get("n", 10)
        title = self.settings.get("title", "Top N Bar Chart")
        description = self.settings.get("description", "Top N Bar Chart description")

        filter_children = None
        if column is not None:
            filter_children = generate_filter(
                data[column],
                {"type": "card-settings", "id": self.id, "setting": "column"},
                default_value=column_filter,
            )

        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "column",
                    },
                    label="Column",
                    value=column,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "column-filter",
                    },
                    children=filter_children,
                ),
                dmc.NumberInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "n",
                    },
                    label="Number of entries",
                    value=n,
                    min=1,
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )


class HightlightCard(Card):
    title = "Highlight"
    description = "This card shows a highlight of a given dataset"
    icon = "mdi:file-document-edit"

    def render(self):
        column = self.settings.get("column", None)
        aggregation = self.settings.get("aggregation", "count")
        filter_value = self.settings.get("column-filter", None)
        filtered_data = data
        if filter_value is not None:
            if filtered_data[column].dtype in ["object", "string", "bool", "category"]:
                filtered_data = filtered_data[filtered_data[column].isin(filter_value)]
            else:
                filter_value = [float(x) for x in filter_value]
                filtered_data = filtered_data[
                    (filtered_data[column] >= filter_value[0])
                    & (filtered_data[column] <= filter_value[1])
                ]
        highlight_value = filtered_data[column].agg(aggregation)
        if isinstance(highlight_value, float):
            highlight_value = round(highlight_value, 2)
        icon = self.settings.get("icon", "mdi:star")
        suffix = self.settings.get("suffix", "Suffix")
        return (
            dmc.Card(
                [
                    dmc.Group(
                        children=[
                            dmc.Text(
                                suffix,
                                c="dimmed",
                                fz="14px",
                                fw=400,
                            )
                        ],
                        # justify="flex-end",
                    ),
                    dmc.Group(
                        children=[
                            dmc.Text(highlight_value, fz="40px", fw=600, c="blue"),
                            dmc.ThemeIcon(
                                DashIconify(icon=icon, width=50),
                                size=50,
                                radius="xl",
                                variant="light",
                            ),
                        ],
                        justify="space-between",
                        wrap="nowrap",
                    ),
                ],
                style={"height": "100%"},
                withBorder=True,
                shadow="xs",
            ),
        )

    def render_settings(self):
        column = self.settings.get("column", None)
        column_filter = self.settings.get("column-filter", None)
        aggregation = self.settings.get("aggregation", "count")
        suffix = self.settings.get("suffix", "Suffix")
        icon = self.settings.get("icon", "mdi:star")

        filter_children = None
        if column is not None:
            filter_children = generate_filter(
                data[column],
                {"type": "card-settings", "id": self.id, "setting": "column"},
                default_value=column_filter,
            )
        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "column",
                    },
                    label="Column",
                    value=column,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                html.Div(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "container": "column-filter",
                    },
                    children=filter_children,
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "aggregation",
                    },
                    label="Aggregation",
                    value=aggregation,
                    searchable=True,
                    data=[
                        {"label": "Count", "value": "count"},
                        {"label": "Count-Unique", "value": "nunique"},
                        {"label": "Mean", "value": "mean"},
                        {"label": "Sum", "value": "sum"},
                        {"label": "Min", "value": "min"},
                        {"label": "Max", "value": "max"},
                    ],
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "suffix",
                    },
                    label="Suffix",
                    value=suffix,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "icon",
                    },
                    label="Icon",
                    value=icon,
                ),
                html.A(
                    "Icon list",
                    href="https://icon-sets.iconify.design/mdi/?keyword=mdi",
                    target="_blank",
                ),
            ]
        )
    @callback(
        Output(
            {"type": "card-settings", "id": MATCH, "container": "column-filter"}, "children"
        ),
        Input({"type": "card-settings", "id": MATCH, "setting": "column"}, "value"),
    )
    def update_filter_x(value):
        """If the column is categorical, show a dropdown to filter the data
        else if data is numeric, show a slider to filter the data"""
        if value is None:
            return no_update
        column = data[value]
        # get the input id
        ctx = callback_context
        if not ctx.triggered_id:
            return no_update
        input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
        return generate_filter(column, input_id)

class MarkdownCard(Card):
    title = "Markdown"
    description = "This card shows a markdown content"
    icon = "mdi:file-document-edit"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        markdown = self.settings.get("markdown", "### Markdown")
        return dmc.Card(
            dmc.ScrollArea(
                [
                    dcc.Markdown(markdown),
                ],
            ),
            style={"height": "100%"},
            withBorder=True,
        )

    def render_settings(self):
        markdown = self.settings.get("markdown", "### Markdown")
        return dmc.Stack(
            [
                dmc.Textarea(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "markdown",
                    },
                    label="Markdown",
                    value=markdown,
                    autosize=True,
                ),
            ]
        )

class MapCard(Card):
    title = "Map"
    description = "This card shows a map of a given dataset"
    icon = "mdi:map"
    grid_settings = {"w": 4, "h": 2, "minW": 4, "minH": 2}

    def render(self):
        location = self.settings.get("location", None)
        location_mode = self.settings.get("location_mode", "country names")
        value = self.settings.get("value", None)
        aggregation = self.settings.get("aggregation", "sum")
        title = self.settings.get("title", "Map")
        description = self.settings.get(
            "description", f"Map of {location} aggregated by {value}"
        )

        if location and value:
            aggregated_data = (
                data.groupby(location)[value].agg(aggregation).reset_index()
            )
        else:
            aggregated_data = pd.DataFrame()

        figure = px.choropleth(
            aggregated_data,
            locations=location,
            locationmode=location_mode,
            color=value,
            hover_name=location,
            color_continuous_scale=px.colors.sequential.Plasma,
            template="mantine_light",
        )

        if location_mode == "USA-states":
            figure.update_geos(fitbounds="locations", visible=False)
        else:
            figure.update_geos(projection_type="natural earth")

        figure.update_layout(margin=dict(l=0, r=0, t=15, b=0))
        return dmc.Card(
            [
                dmc.Text(title, fz="30px", fw=600, c="blue"),
                dmc.Text(
                    description,
                    fw=600,
                    c="dimmed",
                ),
                dcc.Graph(
                    figure=figure,
                    id={"type": "card-control", "sub-type": "figure", "id": self.id},
                    className="no-drag",
                    responsive=True,
                    style={"height": "100%"},
                ),
            ],
            style={"height": "100%"},
            withBorder=True,
            shadow="xs",
        )

    def render_settings(self):
        location = self.settings.get("location", None)
        location_mode = self.settings.get("location_mode", "country names")
        value = self.settings.get("value", None)
        aggregation = self.settings.get("aggregation", "sum")
        title = self.settings.get("title", "Map")
        description = self.settings.get("description", "Map description")
        return dmc.Stack(
            [
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "location_mode",
                    },
                    label="Location Mode",
                    value=location_mode,
                    data=[
                        {"label": "Country Names", "value": "country names"},
                        {"label": "ISO-3", "value": "ISO-3"},
                        {"label": "USA-states", "value": "USA-states"},
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "location",
                    },
                    label="Location",
                    value=location,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "value",
                    },
                    label="Value",
                    value=value,
                    searchable=True,
                    data=[
                        {"label": column, "value": column} for column in data.columns
                    ],
                ),
                dmc.Select(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "aggregation",
                    },
                    label="Aggregation",
                    value=aggregation,
                    data=[
                        {"label": "Sum", "value": "sum"},
                        {"label": "Mean", "value": "mean"},
                        {"label": "Count", "value": "count"},
                        {"label": "Count-Unique", "value": "nunique"},
                        {"label": "Min", "value": "min"},
                        {"label": "Max", "value": "max"},
                    ],
                ),
                dmc.TextInput(
                    id={"type": "card-settings", "id": self.id, "setting": "title"},
                    label="Title",
                    value=title,
                ),
                dmc.TextInput(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "description",
                    },
                    label="Description",
                    value=description,
                ),
            ]
        )

@callback(
    Output({"type": "card-control", "sub-type": "figure", "id": ALL}, "figure"),
    Input("mantine-provider", "forceColorScheme"),
    State({"type": "card-control", "sub-type": "figure", "id": ALL}, "id"),
)
def update_color_scheme(color_scheme, figure_ids):
    template = (
        pio.templates["mantine_light"]
        if color_scheme == "light"
        else pio.templates["mantine_dark"]
    )
    patched_figures = []
    for _ in figure_ids:
        patched_figure = Patch()
        patched_figure["layout"]["template"] = template
        patched_figures.append(patched_figure)
    return patched_figures


canvas = CardCanvas(settings)
canvas.card_manager.register_card_class(HistogramCard)
canvas.card_manager.register_card_class(HeatMap)
canvas.card_manager.register_card_class(ViolinCard)
canvas.card_manager.register_card_class(HightlightCard)
canvas.card_manager.register_card_class(BarChartCard)
canvas.card_manager.register_card_class(MarkdownCard)
canvas.card_manager.register_card_class(TopNBarChartCard)
canvas.card_manager.register_card_class(MapCard)
server = canvas.app.server

if __name__ == "__main__":
    canvas.app.run_server(debug=True)



from cardcanvas import CardCanvas, Card, GlobalSettings
import json
import dash_mantine_components as dmc
import datetime

class Settings(GlobalSettings):
    def render_settings(self):
        return dmc.Stack(
            [
                dmc.Text("Some setting"),
                dmc.Select(
                    id={"type": "global-settings", "setting": "global_setting_1"},
                    value=self.settings.get("global_setting_1", "Option 1"),
                    data=[
                        {"label": "Option 1", "value": "Option 1"},
                        {"label": "Option 2", "value": "Option 2"},
                        {"label": "Option 3", "value": "Option 3"},
                        {"label": "Option 4", "value": "Option 4"},
                    ],
                ),
            ]
        )


settings = {
    "title": "Card Canvas Demo",
    "subtitle": "A Demo application showing the capabilities of CardCanvas",
    "start_config": {},
    "logo": "https://img.icons8.com/?size=80&id=cjlQopC5NR3D&format=png",
    "grid_compact_type": "vertical",
    "grid_row_height": 100,
    "grid_cols": {"xl": 24, "lg": 18, "md": 12, "sm": 6, "xs": 4, "xxs": 2},
    "grid_breakpoints": {
        "xl": 1600,
        "lg": 1200,
        "md": 960,
        "sm": 600,
        "xs": 480,
        "xxs": 320,
    },
    "background_color": "light-dark(#eee, #223)",
    "footer_component": dmc.Card(
        dmc.Box("This is the footer of the Card Canvas demo application"),
    ),
}


swatches = [
    "#25262b",
    "#868e96",
    "#fa5252",
    "#e64980",
    "#be4bdb",
    "#7950f2",
    "#4c6ef5",
    "#228be6",
    "#15aabf",
    "#12b886",
    "#40c057",
    "#82c91e",
    "#fab005",
    "#fd7e14",
]


class TimeCard(Card):
    title = "Display Time"
    description = "Display current time on the card and update every minute"
    icon = "mdi:clock"
    color = "blue"
    interval = 1000 * 60
    grid_settings = {"w": 6, "minW": 6}

    def render(self):
        """Time card render"""
        return dmc.Card(
            [
                dmc.Title(
                    f"Now time is: {datetime.datetime.now().strftime('%H:%M:%S')}",
                    c=self.settings.get("text-color", "white"),
                    order=2,
                ),
            ],
            style={
                "height": "100%",
                "width": "100%",
                "background": self.settings.get("background-color", "#336699"),
            },
        )

    def render_settings(self):
        return dmc.Stack(
            [
                dmc.Text("Text Color"),
                dmc.ColorPicker(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "text-color",
                    },
                    value=self.settings.get("text-color", "white"),
                    swatches=swatches,
                ),
                dmc.Text("Background Color"),
                dmc.ColorPicker(
                    id={
                        "type": "card-settings",
                        "id": self.id,
                        "setting": "background-color",
                    },
                    value=self.settings.get("background-color", "#336699"),
                    swatches=swatches,
                ),
            ]
        )

class GlobalSettingsCard(Card):
    title = "Global Settings"
    description = "Display the global settings"
    icon = "mdi:settings"
    color = "purple"

    def render(self):
        """Global settings card render"""
        return dmc.Card(
            [
                dmc.JsonInput(
                    value=json.dumps(self.global_settings, indent=2),
                    minRows=15,
                    maxRows=15,
                    style={"width": "100%", "height": "100%"},
                    resize="vertical",
                    className="no-drag",
                ),
            ],
            h="100%",
            withBorder=True,
        )

    def render_settings(self):
        return dmc.Stack(
            [
                dmc.Text(
                    "This card does not have any settings, it just displays the global settings"
                ),
            ]
        )

class Options(Card):
    title = "List of options"
    description = "Select from a list of options"
    icon = "mdi:form-select"
    color = "green"

    def render(self):
        """Options card render"""
        return dmc.Card(
            [
                dmc.Text(
                    (
                        f"You have selected {','.join(self.settings.get('option', []))}"
                        f" and the global_setting_1 is {self.global_settings.get('global_setting_1', 'not set')}"
                    ),
                ),
            ],
            style={"height": "100%", "width": "100%"},
            withBorder=True,
        )

    def render_settings(self):
        return dmc.Stack(
            [
                dmc.MultiSelect(
                    id={"type": "card-settings", "id": self.id, "setting": "option"},
                    placeholder="Select an option",
                    label="Select an option",
                    value=self.settings.get("option", []),
                    data=[
                        {"label": "Option 1", "value": "option1"},
                        {"label": "Option 2", "value": "option2"},
                        {"label": "Option 3", "value": "option3"},
                    ],
                ),
            ]
        )


class ColorCard(Card):
    title = "Color Card"
    description = "This card just shows a coloured background"
    icon = "mdi:color"
    color = "orange"

    def render(self):
        """Color card render"""
        return dmc.Paper(
            [dmc.Card(bg=self.settings.get("color", "orange"), h="100%")],
            h="100%",
        )

    def render_settings(self):
        return dmc.Stack(
            [
                dmc.ColorPicker(
                    id={"type": "card-settings", "id": self.id, "setting": "color"},
                    value=self.settings.get("color", "grey"),
                    swatches=swatches,
                ),
            ]
        )


canvas = CardCanvas(settings)
canvas.card_manager.register_card_class(TimeCard)
canvas.card_manager.register_card_class(GlobalSettingsCard)
canvas.card_manager.register_card_class(ColorCard)
canvas.card_manager.register_card_class(Options)
canvas.card_manager.register_global_settings_class(Settings)

canvas.app.run(debug=True)



import pandas as pd

nea_data = pd.read_csv(
    "https://raw.githubusercontent.com/plotly/Figure-Friday/refs/heads/main/2025/week-4/Post45_NEAData_Final.csv"
)
nea_data["Age"] = nea_data["nea_grant_year"] - nea_data["birth_year"]
